"""
Initialize Bluemist-AI's environment
"""

# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: Feb 10, 2023
# Last modified: June 22, 2023

import logging
import os
import platform
import shutil
import sysconfig
import traceback
from logging import config

from termcolor import colored

from bluemist.utils.constants import CPU_BRAND_INTEL, GPU_BRAND_INTEL, GPU_BRAND_NVIDIA

os.environ["BLUEMIST_PATH"] = os.path.realpath(os.path.dirname(__file__))
BLUEMIST_PATH = os.getenv("BLUEMIST_PATH")

os.chdir(BLUEMIST_PATH)

config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")
logging.captureWarnings(True)
logger.info('BLUEMIST_PATH {}'.format(BLUEMIST_PATH))

available_gpu = None
available_cpu = None


def initialize(
        log_level='DEBUG',
        enable_acceleration_extensions=False,
        cleanup_resources=True
):
    """
    log_level : {'CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'WARN', 'INFO', 'DEBUG'}, default='DEBUG'
        Controls the logging level for bluemist.log
    enable_acceleration_extensions : {True, False}, default=False
        - Enables NVIDIA GPU acceleration/Intel CPU acceleration based on the underlying GPU/CPU infrastructure
        - NVIDIA GPU acceleration is provided by RAPIDS cuML. For the list of supported algorithms, please refer  https://docs.rapids.ai/api/cuml/stable/api/#regression-and-classification
        - Intel CPU acceleration is provided by Intel® Extension for Scikit-learn. For the list of supported algorithms, please refer https://intel.github.io/scikit-learn-intelex/algorithms.html#on-cpu
    cleanup_resources : {True, False}, default=True
        Cleanup artifacts from previous runs
    """

    global available_gpu
    global available_cpu

    banner = """
    ██████╗ ██╗     ██╗   ██╗███████╗███╗   ███╗██╗███████╗████████╗     █████╗ ██╗
    ██╔══██╗██║     ██║   ██║██╔════╝████╗ ████║██║██╔════╝╚══██╔══╝    ██╔══██╗██║
    ██████╔╝██║     ██║   ██║█████╗  ██╔████╔██║██║███████╗   ██║       ███████║██║
    ██╔══██╗██║     ██║   ██║██╔══╝  ██║╚██╔╝██║██║╚════██║   ██║       ██╔══██║██║
    ██████╔╝███████╗╚██████╔╝███████╗██║ ╚═╝ ██║██║███████║   ██║       ██║  ██║██║                                                                        
                        (version 0.1.2 - ThunderBurst)
    """

    print(colored(banner, 'blue'))

    print('Bluemist path :: {}'.format(BLUEMIST_PATH))
    print('System platform :: {}, {}, {}, {}, {}'.format(os.name, platform.system(), platform.release(),
                                                         sysconfig.get_platform(),
                                                         platform.architecture()))

    logger.info('System platform :: {}, {}, {}, {}, {}'.format(os.name, platform.system(), platform.release(),
                                                               sysconfig.get_platform(),
                                                               platform.architecture()))

    logger.debug('Printing environment variables...')

    for key, value in os.environ.items():
        logger.debug(f'{key}={value}')

    if log_level.upper() in ['CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'WARN', 'INFO', 'DEBUG']:
        logger.setLevel(logging.getLevelName(log_level))

    logger.handlers[0].doRollover()

    if enable_acceleration_extensions:
        gpu_brand = check_gpu_brand()
        cpu_brand = check_cpu_brand()

        print("GPU Brand ::", gpu_brand)
        print("CPU Brand ::", cpu_brand)

        if gpu_brand == GPU_BRAND_NVIDIA:
            try:
                import cuml
                cuml_version = cuml.__version__
                available_gpu = gpu_brand
                print("cuML version", str(cuml_version))
                print("NVIDIA GPU support is available via RAPIDS cuML ", str(cuml_version))
            except Exception as e:
                print("NVIDIA GPU support is NOT available !")
                logger.error("Error: %s", str(e))
                logger.error(traceback.format_exc())

        if cpu_brand == CPU_BRAND_INTEL:
            try:
                from sklearnex import patch_sklearn
                patch_sklearn()
                available_cpu = cpu_brand
                print("CPU Acceleration enabled via Intel® Extension for Scikit-learn")
            except Exception as e:
                print("Intel CPU Acceleration is NOT available !")
                logger.error("Error: %s", str(e))
                logger.error(traceback.format_exc())

    # cleaning and building artifacts directory
    if bool(cleanup_resources):
        directories = ['artifacts/data', 'artifacts/eda', 'artifacts/experiments', 'artifacts/models',
                       'artifacts/preprocessor', 'mlruns']
        for directory in directories:
            directory_path = BLUEMIST_PATH + '/' + directory
            if os.path.exists(directory_path):
                logger.debug('Removing directory :: {}'.format(directory_path))
                shutil.rmtree(directory_path)
            if not os.path.exists(directory_path):
                logger.debug('Creating directory :: {}'.format(directory_path))
                os.mkdir(directory_path)

        with open(BLUEMIST_PATH + '/' + 'artifacts/api/predict.py', 'w') as f:
            logger.debug('Clearing file content of {}'.format(BLUEMIST_PATH + '/' + 'artifacts/api/predict.py'))
            f.truncate()


def check_gpu_brand():
    import subprocess

    # Check for NVIDIA GPU
    try:
        subprocess.check_output(['nvidia-smi', '--help'])
        return GPU_BRAND_NVIDIA
    except (FileNotFoundError, subprocess.CalledProcessError):
        pass

    # Check for Intel GPU
    try:
        subprocess.check_output(['intel_gpu_top', '-h'])
        return GPU_BRAND_INTEL
    except (FileNotFoundError, subprocess.CalledProcessError):
        pass

    return "Unknown GPU brand !!"


def check_cpu_brand():
    import cpuinfo

    cpu_brand = cpuinfo.get_cpu_info()['vendor_id_raw']  # get only the brand name
    return cpu_brand




__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"


from bluemist.preprocessing.preprocessor import (preprocess_data)

__all__ = [
    'preprocess_data'
]



"""
Performs model training, testing, evaluations and deployment
"""

# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created:  Jun 22, 2022
# Last modified: June 19, 2023

import importlib
import logging
import os
import time
from logging import config
import pandas as pd

from pandas.core.dtypes.common import is_numeric_dtype
from sklearn.compose import TransformedTargetRegressor
from sklearn.model_selection import RandomizedSearchCV
from sklearn.pipeline import Pipeline
from tqdm import tqdm

import bluemist
from bluemist import environment
from bluemist.pipeline.bluemist_pipeline import add_pipeline_step, save_model_pipeline, clear_all_model_pipelines
from bluemist.preprocessing import preprocessor
from bluemist.regression.constant import multi_output_regressors, multi_task_regressors, unsupported_regressors, \
    base_estimator_regressors
from bluemist.utils.constants import GPU_BRAND_INTEL, GPU_BRAND_NVIDIA, GPU_ACCELERATION_NVIDIA, CPU_ACCELERATION_INTEL, \
    CPU_BRAND_INTEL

from bluemist.utils.metrics import metric_scorer
from sklearn.utils import all_estimators

from bluemist.utils.scaler import get_scaler
from bluemist.utils import generate_api as generate_api
from bluemist.artifacts.api import predict

from IPython.display import display, HTML

BLUEMIST_PATH = os.environ["BLUEMIST_PATH"]

config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")


def import_mlflow():
    import mlflow
    return mlflow


def measure_execution_time(start_time):
    # Calculate the elapsed time
    elapsed_time = time.time() - start_time

    # Convert elapsed time to seconds, minutes, milliseconds, and hours
    milliseconds = int(elapsed_time * 1000) % 1000
    seconds = elapsed_time % 60
    minutes = (elapsed_time // 60) % 60
    hours = (elapsed_time // 3600)

    execution_time = "{:02d}:{:02d}:{:02d},{:03d}".format(int(hours), int(minutes), int(seconds), int(milliseconds))
    return execution_time


def initialize_mlflow(mlflow_experiment_name):
    mlflow = import_mlflow()

    logger.info('Initializing MLFlow...')
    if mlflow_experiment_name is not None:
        experiment = mlflow.get_experiment_by_name(mlflow_experiment_name)
        if not experiment:
            mlflow.create_experiment(name=mlflow_experiment_name,
                                     artifact_location=BLUEMIST_PATH + '/' + 'artifacts/experiments/mlflow')

        if experiment is not None and experiment.lifecycle_stage == 'deleted':
            logger.info('Restoring MLFlow experiment :: {}'.format(mlflow_experiment_name))
            client = mlflow.tracking.MlflowClient()
            client.restore_experiment(experiment.experiment_id)

        mlflow.set_experiment(mlflow_experiment_name)


def get_estimators(multi_output=False, multi_task=False, names_only=True):
    """
        **Returns the list of available regression estimators**

        multi_output : bool, default=False
            Future use
        multi_task : bool, default=False
            Future use
        names_only : bool, default=True
            Returns only the estimator name without metadata
    """

    estimators = all_estimators(type_filter='regressor')
    logger.debug('All available estimators :: {}'.format(estimators))

    estimators_to_remove = []
    for estimator in estimators:
        if not multi_output and estimator[0] in multi_output_regressors:
            estimators_to_remove.append(estimator)
        if not multi_task and estimator[0] in multi_task_regressors:
            estimators_to_remove.append(estimator)
        if unsupported_regressors and estimator[0] in unsupported_regressors:
            estimators_to_remove.append(estimator)
        if base_estimator_regressors and estimator[0] in base_estimator_regressors:
            estimators_to_remove.append(estimator)

    logger.debug('Estimators not supported by Bluemist AI :: {}'.format(estimators_to_remove))

    for estimator_to_remove in estimators_to_remove:
        estimators.remove(estimator_to_remove)

    if bool(names_only):
        return [estimator[0] for estimator in estimators]

    logger.info('Estimators available for modelling :: {}'.format(estimators))
    return estimators


def deploy_model(estimator_name, host='localhost', port=8000):
    """
        estimator_name : str
            Name of the estimator to be deployed
        host : {str, IPv4 or IPv6}, default='localhost'
            Hostname or IP address of the machine where the API will be deployed
        port : int, default=8000
            API listening port
    """

    logger.info('Generating API code to deploy the model :: {}'.format(estimator_name))
    generate_api.generate_api_code(estimator_name=estimator_name,
                                   initial_column_metadata=preprocessor.initial_column_metadata_for_deployment,
                                   encoded_column_metadata=preprocessor.encoded_columns_for_deployment,
                                   target_variable=preprocessor.target_for_deployment)
    importlib.reload(predict)
    logger.info('Starting API server on host {} and port {}'.format(host, port))
    predict.start_api_server(host=host, port=port)


def train_test_evaluate(
        X_train,
        X_test,
        y_train,
        y_test,
        tune_models=None,
        metrics='default',
        multi_output=False,
        multi_task=False,
        target_scaling_strategy=None,
        save_pipeline_to_disk=True,
        experiment_name=None,
        run_name=None):
    """
        **Trains the data on the given dataset, evaluate the models and returns comparison metrics**

        X_train : pandas dataframe
            Training data
        X_test : pandas dataframe
            Test data
        y_train : array of shape (X_train.shape[0],)
            Target values of training dataset
        y_test : array of shape (X_test.shape[0],)
            Target values of test dataset
        tune_models : {'all', None} or list of models to be trained, default=None
            all: tune all regression models
            list: list of models to be trained
            None: hyperparameter tuning will not be performed
        metrics : {'all', 'default'}, default='default'
            - all:
                mean_absolute_error, mean_squared_error, r2_score, explained_variance_score, max_error,
                mean_squared_log_error, median_absolute_error, mean_absolute_percentage_error, mean_poisson_deviance,
                mean_gamma_deviance, mean_tweedie_deviance, d2_tweedie_score, mean_pinball_loss
            - default:
                mean_absolute_error, mean_squared_error, r2_score
        multi_output : bool, default=False
            Future use
        multi_task : bool, default=False
            Future use
        target_scaling_strategy : {'StandardScaler', 'MinMaxScaler', 'MaxAbsScaler', 'RobustScaler', None}, default=None
            Scales the target variable before training the model
        save_pipeline_to_disk : bool, default=True
            Save preprocessor and model training pipeline to the disk. Should be set to True if needs model to be deployed
            as an API
        experiment_name : str, default=None
            Name of the experiment
        run_name : str,default=None
            Name of the run within the experiment

        Examples
        ---------
        *Regression*

        .. raw:: html
            :file: ../../code_samples/quickstarts/regression/regression_hyperparameter_tuning.html

    """

    tune_all_models = False
    models_to_tune = []
    target_scaler = None
    sklearnex_algorithms = None

    if target_scaling_strategy is not None:
        target_scaler = get_scaler(target_scaling_strategy)

    if isinstance(tune_models, str) and tune_models == 'all':
        tune_all_models = True
    elif isinstance(tune_models, list):
        models_to_tune = tune_models

    if experiment_name is not None:
        initialize_mlflow(experiment_name)

    # Get patch names from sklearnex
    if environment.available_cpu == CPU_BRAND_INTEL:
        from sklearnex import sklearn_is_patched, get_patch_names
        sklearnex_algorithms = get_patch_names()

    df = pd.DataFrame()

    estimators = get_estimators(multi_output, multi_task, names_only=False)
    clear_all_model_pipelines()

    counter = 0

    # If hyperparameter tuning is requested for specific models, limit the overall training to those models to save time
    if tune_models is not None and not tune_all_models and models_to_tune:
        estimators_to_skip = []
        for estimator in estimators:
            if estimator[0] not in models_to_tune:
                estimators_to_skip.append(estimator)

        for estimator_to_skip in estimators_to_skip:
            estimators.remove(estimator_to_skip)

    for estimator_name, estimator_class in (pbar := tqdm(estimators, colour='blue')):
        start_time = time.time()
        pbar.set_description(f"Training {estimator_name}")

        counter = counter + 1
        estimator_execution_device = 'CPU'

        # No tuning OR Tune All OR Tune Few
        if tune_models is None or tune_all_models or estimator_name in models_to_tune:
            logger.info('############  Regressor in progress :: {} ############'.format(estimator_name))
            regressor = None

            try:
                if environment.available_gpu == GPU_BRAND_NVIDIA:
                    import cuml
                    cuml.set_global_output_type('array')  # cuML will return predictions of type cupy.ndarray
                    if hasattr(cuml, estimator_name):
                        regressor = getattr(cuml, estimator_name)()
                        estimator_execution_device = GPU_ACCELERATION_NVIDIA
                        logger.info('Regressor class from cuML :: {}'.format(regressor.__class__.__module__ + '.' + regressor.__class__.__name__))

                if regressor is None and environment.available_cpu == CPU_BRAND_INTEL:
                    for sklearnex_algorithm in sklearnex_algorithms:
                        if (estimator_name == 'LinearRegression' and sklearnex_algorithm == 'linear') or estimator_name.lower() == sklearnex_algorithm:
                            logger.info('\nSklearn Algorithm :: {}, Sklearn Intel(R) Ex Algorithm :: {}'.format(estimator_name, sklearnex_algorithm))
                            if sklearn_is_patched(name=sklearnex_algorithm):
                                regressor = estimator_class()
                                estimator_execution_device = CPU_ACCELERATION_INTEL
                            break

                # Normal CPU processing if acceleration extensions are not applied
                if regressor is None:
                    regressor = estimator_class()

                # TODO: Revisit this code. sklearn throws error without n_components=1
                if estimator_name in ['CCA', 'PLSCanonical']:
                    regressor.set_params(n_components=1)

                # Hyperparameter tuning is requested
                if tune_all_models or estimator_name in models_to_tune:
                    estimator_params = regressor.get_params()

                    logger.info('Available hyperparameters to be tuned :: {}'.format(estimator_params))
                    logger.debug('Python type() for hyperparameters :: {}'.format(type(estimator_params)))

                    if estimator_execution_device == GPU_ACCELERATION_NVIDIA:
                        default_hyperparameters_module = bluemist.regression.tuning.cuml
                    else:
                        default_hyperparameters_module = bluemist.regression.tuning.sklearn

                    default_hyperparameters = getattr(default_hyperparameters_module, 'default_params', None)
                    model_params_for_tuning = getattr(default_hyperparameters_module, estimator_name, None)
                    remove_params = model_params_for_tuning.get('remove_params')

                    hyperparameters_to_remove = []
                    for hyperparameter, default_value in estimator_params.items():
                        if str(default_value) == 'deprecated':
                            hyperparameters_to_remove.append(hyperparameter)
                            logger.debug('Deprecated hyperparameter identified :: {}'.format(hyperparameter))
                        elif model_params_for_tuning is not None and hyperparameter in model_params_for_tuning:
                            estimator_params[hyperparameter] = model_params_for_tuning[hyperparameter]
                            logger.debug('Hyperparameter in model configuration :: {} :: {}'.format(hyperparameter, estimator_params[hyperparameter]))
                        elif hyperparameter in default_hyperparameters:
                            estimator_params[hyperparameter] = default_hyperparameters[hyperparameter]
                            logger.debug('Hyperparameter in default configuration :: {} :: {}'.format(hyperparameter, estimator_params[hyperparameter]))

                        if remove_params is not None:
                            if hyperparameter in remove_params and hyperparameter not in hyperparameters_to_remove:
                                hyperparameters_to_remove.append(hyperparameter)
                                logger.debug('Unsupported hyperparameter identified :: {}'.format(hyperparameter))

                    for hyperparameter_to_remove in hyperparameters_to_remove:
                        estimator_params.pop(hyperparameter_to_remove, None)

                    # Creating new dictionary of hyperparameters to add step name as required by the sklearn pipeline
                    hyperparameters = {}
                    for hyperparameter in estimator_params:
                        old_key = hyperparameter

                        if target_scaling_strategy is not None:
                            new_key = estimator_name + '__regressor__' + hyperparameter
                        else:
                            new_key = estimator_name + '__' + hyperparameter
                        hyperparameters[new_key] = estimator_params[old_key]

                    logger.info('Hyperparameters to be used for model tuning :: {}'.format(hyperparameters))

                    if target_scaling_strategy is not None:
                        # TODO: Remove the usage of TransformedTargetRegressor so speical handling is not required for cuML
                        transformed_target_regressor = TransformedTargetRegressor(regressor=regressor, transformer=target_scaler)

                        if environment.available_gpu == 'NVIDIA':
                            step_estimator = (estimator_name, regressor)
                        else:
                            step_estimator = (estimator_name, transformed_target_regressor)
                        steps = add_pipeline_step(estimator_name, step_estimator)
                    else:
                        step_estimator = (estimator_name, regressor)
                        steps = add_pipeline_step(estimator_name, step_estimator)

                    if steps is not None:
                        model_pipeline = Pipeline(steps=steps)
                        randomized_search = RandomizedSearchCV(model_pipeline, param_distributions=hyperparameters, n_iter=100, error_score='raise')
                        optimized_estimator = randomized_search.fit(X_train, y_train)
                        best_estimator_pipeline = optimized_estimator.best_estimator_

                        logger.debug('Model pipeline parameters :: {}'.format(model_pipeline.get_params().keys()))
                        logger.info('Fitted estimator with all parameters :: {}'.format(optimized_estimator))
                        logger.debug('Model pipeline :: {}'.format(model_pipeline))
                        logger.info('Model pipeline with best estimator :: {}'.format(best_estimator_pipeline))

                        if save_pipeline_to_disk:
                            logger.info('Saving model pipeline to disk')
                            save_model_pipeline(estimator_name, best_estimator_pipeline)
                else:
                    if target_scaling_strategy is not None:
                        # TODO: Remove the usage of TransformedTargetRegressor so speical handling is not required for cuML
                        transformed_target_regressor = TransformedTargetRegressor(regressor=regressor, transformer=target_scaler)

                        if environment.available_gpu == 'NVIDIA':
                            step_estimator = (estimator_name, regressor)
                        else:
                            step_estimator = (estimator_name, transformed_target_regressor)
                        steps = add_pipeline_step(estimator_name, step_estimator)
                    else:
                        step_estimator = (estimator_name, regressor)
                        steps = add_pipeline_step(estimator_name, step_estimator)

                    model_pipeline = Pipeline(steps=steps)
                    best_estimator_pipeline = model_pipeline.fit(X_train, y_train)

                    if save_pipeline_to_disk:
                        save_model_pipeline(estimator_name, best_estimator_pipeline)

                    logger.info('Model pipeline with best estimator (no hyperparameter tuning) :: {}'.format(best_estimator_pipeline))
                    logger.debug('Model pipeline (no hyperparameter tuning) :: {}'.format(model_pipeline))

                if tune_all_models or estimator_name in models_to_tune:
                    logger.info('Best score :: {}'.format(optimized_estimator.best_score_))
                    logger.info('Best hyperparameters :: {}'.format(optimized_estimator.best_params_))

                y_pred = best_estimator_pipeline.predict(X_test)

                # Convert cupy.ndarray to numpy.ndarray as sklearn returns numpy.ndarray but cuML will return cupy.ndarray
                y_pred_class_name = y_pred.__class__.__module__ + '.' + y_pred.__class__.__name__
                logger.info('y_pred_class_name :: {}'.format(y_pred_class_name))
                if y_pred_class_name == 'cupy.ndarray':
                    import cupy as cp
                    y_pred = cp.asnumpy(y_pred)

                scorer = metric_scorer(y_test, y_pred, metrics)
                estimator_stats_df = scorer.calculate_metrics()

                final_stats_df = estimator_stats_df.copy()
                execution_time = measure_execution_time(start_time)

                final_stats_df.insert(0, 'Estimator', estimator_name)  # Insert Estimator name as the first column in the dataframe
                final_stats_df.insert(1, 'Execution Device', estimator_execution_device)  # Insert Execution Device as the second column in the dataframe
                final_stats_df.insert(2, 'Execution Time', execution_time) # Insert Execution Time as the third column in the dataframe

                logger.info('Current estimator Stats :: \n{}'.format(final_stats_df.to_string()))
                logger.debug('Current estimator stats as dictionary :: {}'.format(final_stats_df.to_dict('records')))

                df = pd.concat([df, final_stats_df], ignore_index=True)
                logger.debug('Estimator stats so far :: \n{}'.format(df.to_string()))

                if experiment_name is not None:
                    mlflow = import_mlflow()
                    with mlflow.start_run(run_name=run_name):
                        logger.info('Capturing stats in MLFlow...')
                        mlflow.log_param('model', estimator_name)
                        mlflow.log_metrics(estimator_stats_df.to_dict('records')[0])
                        mlflow.sklearn.log_model(best_estimator_pipeline, 'model_' + estimator_name)
                        run_id = mlflow.active_run().info.run_id
                        logger.info('Model saved in run :: {}'.format(run_id))
            except Exception as e:
                exception = {'Estimator': [estimator_name], 'Exception': str(e)}
                exception_df = pd.DataFrame(exception)
                logger.info('Exception occurred :: \n{}'.format(exception_df))
                df = pd.concat([df, exception_df])
                logger.error('Exception occurred while training the model :: {}'.format(str(e)), exc_info=True)

    df.set_index('Estimator', inplace=True)
    #print(df)
    display(HTML(df.style
                 .highlight_max(
        subset=[col for col in df.columns if col.endswith('score') and is_numeric_dtype(df[col])], color='green')
                 .highlight_min(
        subset=[col for col in df.columns if col.endswith('score') and is_numeric_dtype(df[col])], color='yellow')
                 .highlight_max(
        subset=[col for col in df.columns if not col.endswith('score') and is_numeric_dtype(df[col])], color='yellow')
                 .highlight_min(
        subset=[col for col in df.columns if not col.endswith('score') and is_numeric_dtype(df[col])], color='green')
                 .to_html()))
    logger.info('Estimator stats across all trained models : \n{}'.format(df.to_string()))




# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created:  Jun 22, 2022
# Last modified: June 17, 2023


import logging
import os
from logging import config
import pandas as pd

BLUEMIST_PATH = os.getenv("BLUEMIST_PATH")

config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")


def get_data_from_filesystem(file_path,
                             file_type='delimited',
                             sheet_name=0,
                             delimiter=','):
    """
        Extract data from local file system

        file_path: str
            File system path of the data file to be extracted
        file_type: {'delimited', 'excel'}, default='delimited'
            Type of the data file
        sheet_name: str, default=0
            Sheet name if ``file_type`` is ``excel``
        delimiter: str, default=','
            File delimiter to use if ``file_type`` is ``delimited``

        Examples
        ---------

        .. raw:: html
           :file: ../../code_samples/quickstarts/datasource/ds_file_system.html

    """

    logger.info('Pulling data from {}'.format(file_path))

    if file_type == 'delimited':
        data = pd.read_csv(filepath_or_buffer=file_path, sep=delimiter)
        logger.info('Data pull completed !!')
        data.columns = data.columns.str.replace('\W', '_', regex=True)  # TODO: Revisit this code
        return data
    elif file_type == 'excel':
        data = pd.read_excel(io=file_path, sheet_name=sheet_name)
        logger.info('Data pull completed !!')
        data.columns = data.columns.str.replace('\W', '_', regex=True)  # TODO: Revisit this code
        return data






__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"


import logging
import os
from logging import config

import joblib
from joblib import dump

pipeline_steps = {}
pipelines = {}

BLUEMIST_PATH = os.environ["BLUEMIST_PATH"]
config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")


def save_preprocessor(preprocessor):
    preprocessor_disk_location = BLUEMIST_PATH + '/' + 'artifacts/preprocessor/preprocessor.joblib'
    logger.info('Saving preprocessor to disk on :: {}'.format(preprocessor_disk_location))
    dump(preprocessor, preprocessor_disk_location)
    logger.debug('Preprocessor column transformer object :: {}'.format(preprocessor))


def add_pipeline_step(estimator_name, pipeline_step):
    if estimator_name not in pipeline_steps:
        pipeline_steps[estimator_name] = []

    if estimator_name in pipeline_steps:
        steps = pipeline_steps[estimator_name]
        steps.append(pipeline_step)
        logger.debug('Model pipeline steps :: {}'.format(steps))
        return steps


def save_model_pipeline(estimator_name, pipeline):
    model_pipeline_disk_location = BLUEMIST_PATH + '/' + 'artifacts/models/' + estimator_name + '.joblib'
    pipelines[estimator_name] = pipeline
    logger.info('Saving model pipeline to disk :: {}'.format(model_pipeline_disk_location))
    dump(pipeline, model_pipeline_disk_location)
    logger.info('Model pipeline object :: {}'.format(pipeline))


def clear_all_model_pipelines():
    global pipeline_steps, pipelines
    pipeline_steps = {}
    pipelines = {}


def get_model_pipeline(estimator_name):
    logger.info('Getting model pipeline for {}'.format(estimator_name))
    pipeline = pipelines[estimator_name]
    return pipeline



# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: Jun 22, 2022
# Last modified: May 29, 2023

from bluemist.regression import tuning
from bluemist.regression.regressor import (
    get_estimators,
    train_test_evaluate,
    deploy_model
)

__all__ = ['get_estimators',
           'train_test_evaluate',
           'deploy_model'
]




__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"


multi_output_regressors = [
    'MultiOutputRegressor',
    'RegressorChain'
]

multi_task_regressors = [
    'MultiTaskElasticNet',
    'MultiTaskLasso',
    'MultiTaskElasticNetCV',
    'MultiTaskLassoCV'
]

unsupported_regressors = [
    'IsotonicRegression'
]

base_estimator_regressors = [
    'StackingRegressor',
    'VotingRegressor'
]



__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"

import logging
import os
from logging import config

from bluemist.datasource import get_data_from_filesystem

BLUEMIST_PATH = os.getenv("BLUEMIST_PATH")
config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")


def get_data_from_s3(aws_access_key_id,
                     aws_secret_access_key,
                     s3_bucket_name,
                     s3_object_name,
                     destination_path,
                     file_type='delimited',
                     sheet_name=0,
                     delimiter=','):
    """
        Extract data from Amazon cloud (AWS)

        aws_access_key_id: str
            The access key to use when creating the s3 client
        aws_secret_access_key: str
            The secret key to use when creating the s3 client
        s3_bucket_name: str
            s3 bucket name from where the dat file needs to be pulled
        s3_object_name: str
            Name of the data file
        destination_path: str
            File system path where the file will be downloaded from S3
        file_type: {'delimited', 'excel'}, default='delimited'
            type of the data file
        sheet_name: str, default=0
            sheet name if the dataset is an Excel file
        delimiter: str, default=','
            file delimiter to use for delimited files

        Examples
        ---------

        .. raw:: html
           :file: ../../code_samples/quickstarts/datasource/ds_aws.html

    """
    import boto3
    s3 = boto3.client('s3', aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)
    logger.debug('S3 service client created :: {}'.format(s3))
    s3.download_file(s3_bucket_name, s3_object_name, destination_path)
    data = get_data_from_filesystem(destination_path, sheet_name, file_type, delimiter)
    data.columns = data.columns.str.replace('\W', '_')  # TODO: Revisit the code
    return data



"""
Performs data pre-processing
"""

# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: Jun 22, 2022
# Last modified: June 11, 2023


import logging
import os
from logging import config
import numpy as np
import pandas as pd
import sklearn
from sklearn.compose import ColumnTransformer

from bluemist.pipeline.bluemist_pipeline import save_preprocessor
from bluemist.preprocessing import categorical_transformer, numeric_transformer

BLUEMIST_PATH = os.environ["BLUEMIST_PATH"]

config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")

initial_column_metadata_for_deployment = []
encoded_columns_for_deployment = []
target_for_deployment = None


def preprocess_data(
        data,
        target_variable,
        test_size=0.25,
        data_randomizer=None,
        drop_features=None,
        numerical_features=None,
        force_numeric_conversion=True,
        categorical_features=None,
        convert_values_to_nan=None,
        data_scaling_strategy='StandardScaler',
        data_tranformation_strategy=None,
        missing_values=np.nan,
        numeric_imputer_strategy='mean',
        numeric_constant_value=None,
        categorical_imputer_strategy='most_frequent',
        categorical_constant_value=None,
        categorical_encoder='OneHotEncoder',
        drop_categories_one_hot_encoder=None,
        handle_unknown_one_hot_encoder=None):
    """
        data: pandas daframe
            Dataframe to be processed before passing to the ML estimator
        target_variable: str
            Target variable to be predicted
        test_size: float or int, default=0.25
            Percentage of the data to be used for testing model performance
        data_randomizer: int default=None
            Controls the data split. Provide a value to reproduce the same split.
        drop_features: str ot list
            Drops the features from the dataset
        numerical_features: list, default=None
            Bluemist AI will automatically identify numerical features from the dataset. Provide the list of features to override the type identified by Bluemist AI.
        force_numeric_conversion: bool, default=True
            Gracefully converts features to numeric datatype which are provided under ``numerical_features``
        categorical_features: list, default=None
            Bluemist AI will automatically identify categorical features from the dataset. Provide the list of features to override the type identified by Bluemist.
        convert_to_nan:  str, list, default=None
            Dataset values to be converted to NumPy NaN
        data_scaling_strategy: {None, 'StandardScaler', 'MinMaxScaler', 'MaxAbsScaler', 'RobustScaler'}, default='StandardScaler'
            Scales dataset features, excluding target variable
                - 'StandardScaler':
                - 'MinMaxScaler'
                - 'MaxAbsScaler'
                - 'RobustScaler'
        data_tranformation_strategy: {'box-cox', 'yeo-johnson' or None}, default=None
            Transforms the features, excluding target variable.
                - 'box-cox':
                - 'yeo-johnson':
        missing_values: int, float, str, np.nan, None or pandas.NA, default=np.nan
            All instances of missing_value will be replaced with the user provided imputer strategy
        numeric_imputer_strategy: {'mean, 'median', 'most_frequent', 'constant'}, default='mean'
            Replaces `missing_values` with the strategy provided
        numeric_constant_value: str or number, default=None
            ``numeric_constant_value`` will replace the ``missing_values`` when ``numeric_imputer_strategy`` is passed as ``constant``
        categorical_imputer_strategy:  {'most_frequent', 'constant'}, default='most_frequent'
            Replaces `missing_values` with the strategy provided
        categorical_constant_value: str or number, default=None
            ``categorical_constant_value`` will replace the ``missing_values`` when ``categorical_imputer_strategy`` is passed as ``constant``
        categorical_encoder: {'OneHotEncoder', 'OrdinalEncoder'}, default='OneHotEncoder'
            Encode categorical features
        drop_categories_one_hot_encoder: {‘first’, ‘if_binary’ or None}, default='None'
            Determines strategy to drop one category per feature
                - 'first':
                    drops the first category for each feature.
                - 'if_binary':
                    drops the first category for features with two categories
                - None:
                    Keeps all features and categories
        handle_unknown_one_hot_encoder : {‘error’, ‘ignore’, ‘infrequent_if_exist’}, default=’error’
            Handles unknown category during transform
                - 'error':
                    throws an error if category is unknown
                - 'ignore':
                    ignores if category is unknown, output encoded column for this feature will be all zeroes
                - 'infrequent_if_exist':
                    unknown category will be mapped to infrequent category if exists. If infrequent category does not exist, it
                    will be treated as `ignore`

        Examples
        ---------
        *Data preprocessing :: Categorical Encoder*

        .. raw:: html
           :file: ../../code_samples/quickstarts/preprocessor/preprocessor_categorical.html

    """

    global target_for_deployment
    global initial_column_metadata_for_deployment
    global encoded_columns_for_deployment

    target_for_deployment = target_variable

    logger.info('Shape of the dataset :: {}'.format(data.shape))
    logger.info('Columns in the dataset :: \n{}'.format(data.columns))

    # drop features from the dataset
    if drop_features is not None:
        if isinstance(drop_features, str):
            data.drop([drop_features], axis=1, inplace=True)
        elif isinstance(drop_features, list):
            data.drop(drop_features, axis=1, inplace=True)

    # auto compute numerical and categorical features
    auto_computed_numerical_features = data.select_dtypes(include='number').columns.tolist()
    auto_computed_categorical_features = data.select_dtypes(include='object').columns.tolist()

    final_numerical_features = auto_computed_numerical_features.copy()
    if target_variable in final_numerical_features:
        final_numerical_features.remove(target_variable)

    final_categorical_features = auto_computed_categorical_features.copy()
    if target_variable in final_categorical_features:
        final_categorical_features.remove(target_variable)

    # finalize the list of numerical features
    if auto_computed_numerical_features is not None:
        if numerical_features is not None:
            for numerical_feature in numerical_features:
                if numerical_feature not in auto_computed_numerical_features:
                    final_numerical_features.append(numerical_feature)

        if categorical_features is not None:
            for categorical_feature in categorical_features:
                if categorical_feature in auto_computed_numerical_features:
                    final_numerical_features.remove(categorical_feature)

    # finalize the list of categorical features
    if auto_computed_categorical_features is not None:
        if categorical_features is not None:
            for categorical_feature in categorical_features:
                if categorical_feature not in auto_computed_categorical_features:
                    final_categorical_features.append(categorical_feature)

        if numerical_features is not None:
            for numerical_feature in numerical_features:
                if numerical_feature in auto_computed_categorical_features:
                    final_categorical_features.remove(numerical_feature)

    # prepare final list of columns after preprocessing
    column_list = []
    if bool(final_numerical_features) and bool(final_categorical_features):
        column_list = final_numerical_features + final_categorical_features
    elif bool(final_numerical_features):
        column_list = final_numerical_features
    elif bool(final_categorical_features):
        column_list = final_categorical_features

    # handle non-numeric data in user provided numeric column
    if numerical_features is not None:
        if force_numeric_conversion:
            numeric_conversion_strategy = 'coerce'
        else:
            numeric_conversion_strategy = 'raise'

    logger.debug('data.dtypes before preprocessing  :: \n{}'.format(data.dtypes))

    if numerical_features is not None:
        data[final_numerical_features] = data[final_numerical_features].apply(pd.to_numeric,
                                                                              errors=numeric_conversion_strategy,
                                                                              axis=1)

    data[final_categorical_features] = data[final_categorical_features].astype(str)
    logger.debug('data.dtypes after dtype conversion  :: \n{}'.format(data.dtypes))

    # Creating list of column name and datatype which will be used in generate_api.py
    for col_name, col_type in data.drop(target_variable, axis=1).dtypes.items():
        initial_column_metadata_for_deployment.append((col_name, col_type))

    # create transformers for preprocessing pipeline
    num_transformer = numeric_transformer.build_numeric_transformer_pipeline(**locals())
    cat_transformer = categorical_transformer.build_categorical_transformer_pipeline(**locals())

    # create preprocessing pipeline
    preprocessor = ColumnTransformer(
        transformers=[
            ("numeric_transformer", num_transformer, final_numerical_features),
            ("categorical_transformer", cat_transformer, final_categorical_features)
        ], verbose_feature_names_out=False
    )

    X = data.drop([target_variable], axis=1)
    y = data[[target_variable]]
    logger.debug('Splitting dataset into X_train, X_test, y_train, y_test...')

    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(X, y, test_size=test_size,
                                                                                random_state=data_randomizer)

    logger.debug('X_train.dtypes before ColumnTransformer :: \n{}'.format(X_train.dtypes))
    X_train = pd.DataFrame(preprocessor.fit_transform(X_train), columns=preprocessor.get_feature_names_out())
    X_test = pd.DataFrame(preprocessor.transform(X_test), columns=preprocessor.get_feature_names_out())

    encoded_columns_for_deployment = preprocessor.get_feature_names_out()

    logger.debug('X_train Columns after ColumnTransformer :: {}'.format(preprocessor.get_feature_names_out()))
    logger.debug('X_train.dtypes after ColumnTransformer :: \n{}'.format(X_train.dtypes))

    y_train = np.ravel(y_train)
    y_test = np.ravel(y_test)

    logger.debug('Saving preprocessor to disk...')
    save_preprocessor(preprocessor)
    return X_train, X_test, y_train, y_test



# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created:  Jun 19, 2022
# Last modified: June 19, 2023


import logging
import os
from logging import config
import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, explained_variance_score, max_error, \
    mean_squared_log_error, median_absolute_error, mean_absolute_percentage_error, mean_poisson_deviance, \
    mean_gamma_deviance, mean_tweedie_deviance, d2_tweedie_score, mean_pinball_loss

BLUEMIST_PATH = os.environ["BLUEMIST_PATH"]
config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")

default_regression_metrics = [mean_absolute_error, mean_squared_error, r2_score]
all_regression_metrics = [*default_regression_metrics, explained_variance_score, max_error, mean_squared_log_error,
                          median_absolute_error, mean_absolute_percentage_error, mean_poisson_deviance,
                          mean_gamma_deviance, mean_tweedie_deviance, d2_tweedie_score, mean_pinball_loss]

classification_metrics = []
all_classification_metrics = []


class metric_scorer:
    y_true = None
    y_pred = None
    metrics_requested = None
    metrics_to_be_returned = None

    def __init__(self, y_true, y_pred, metrics):
        self.y_true = y_true
        self.y_pred = y_pred
        self.metrics_requested = metrics
        self.metrics_to_be_returned = []

    def calculate_metrics(self):

        rows = []
        row = []

        logger.info('Metrics requested :: {}'.format(self.metrics_requested))

        indexes_with_nan_predictions = np.argwhere(np.isnan(self.y_pred))
        self.y_true = np.delete(self.y_true, indexes_with_nan_predictions)
        self.y_pred = np.delete(self.y_pred, indexes_with_nan_predictions)
        logger.info('Indexes with NaN predictions to be removed :: {}'.format(indexes_with_nan_predictions))

        if isinstance(self.metrics_requested, str) and self.metrics_requested == 'default':
            for metric in default_regression_metrics:
                self.metrics_to_be_returned.append(metric.__name__)
                try:
                    row.append(metric(self.y_true, self.y_pred))
                except Exception as e:
                    row.append(str(e))
                    logger.error('Exception occurred while computing metric :: {}'.format(str(e)), exc_info=True)
        elif isinstance(self.metrics_requested, str) and self.metrics_requested == 'all':
            for metric in all_regression_metrics:
                self.metrics_to_be_returned.append(metric.__name__)
                try:
                    row.append(metric(self.y_true, self.y_pred))
                except Exception as e:
                    row.append(str(e))
                    logger.error('Exception occurred while computing metric :: {}'.format(str(e)), exc_info=True)
        elif isinstance(self.metrics_requested, list) and len(self.metrics_requested) > 0:
            for metric in self.metrics_requested:
                self.metrics_to_be_returned.append(metric.__name__)
                try:
                    row.append(metric(self.y_true, self.y_pred))
                except Exception as e:
                    row.append(str(e))
                    logger.error('Exception occurred while computing metric :: {}'.format(str(e)), exc_info=True)

        rows.append(row)
        stats_df = pd.DataFrame(rows, columns=self.metrics_to_be_returned)
        return stats_df




__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"


import logging
import os
import urllib.parse
from logging import config
import pandas as pd

BLUEMIST_PATH = os.getenv("BLUEMIST_PATH")

config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")


def get_data_from_database(db_type=None,
                           host=None,
                           database=None,
                           service=None,
                           oracle_instant_client_path=None,
                           username=None,
                           password=None,
                           query=None,
                           chunk_size=1000):
    """
        Extract data from database. Supported databases are MariaDB, Microsoft SQL, MySQL, Oracle, AWS Aurora and PostgreSQL

        db_type: {'mariadb', 'mssql', 'mysql', 'aurora-mysql', 'oracle', 'postgres', 'aurora-postgre'}
            Database type
        host: str
            Database host
        database: str
            Database name
        service: str
            Oracle service name. Used if ``db_type`` is ``oracle``
        oracle_instant_client_path: str
            Filesystem path of Oracle instant client. Used if ``db_type`` is ``oracle``
        username: str
            Database user
        password: str
            Database password
        query: str
            Database query to be used to extract data
        chunk_size: int or None, default=1000
            Number of rows to return in each batch. Pass ``None`` to disable batch mode

        Examples
        ---------

        .. raw:: html
           :file: ../../code_samples/quickstarts/datasource/ds_database.html

    """

    password = urllib.parse.quote_plus(password)

    if db_type == 'mariadb':
        from sqlalchemy import create_engine
        logger.info('Pulling data from MariaDB')
        connection_url = 'mysql+pymysql://' + username + ':' + password + '@' + host + '/' + database
        engine = create_engine(connection_url)
        conn = engine.connect()
        logger.info('Connection successful !!')
        data = extract_data(conn, query, chunk_size)
        return data
    elif db_type == 'mssql':
        from sqlalchemy import create_engine
        logger.info('Pulling data from MS SQL')
        connection_url = 'mssql+pymssql://' + username + ':' + password + '@' + host + '/' + database
        engine = create_engine(connection_url)
        conn = engine.connect()
        logger.info('Connection successful !!')
        data = extract_data(conn, query, chunk_size)
        return data
    elif db_type == 'mysql' or db_type == 'aurora-mysql':
        from sqlalchemy import create_engine
        logger.info('Pulling data from MySQL')
        connection_url = 'mysql+pymysql://' + username + ':' + password + '@' + host + '/' + database
        engine = create_engine(connection_url)
        conn = engine.connect()
        logger.info('Connection successful !!')
        data = extract_data(conn, query, chunk_size)
        return data
    elif db_type == 'oracle':
        from sqlalchemy import create_engine
        import cx_Oracle
        logger.info('Pulling data from Oracle')
        connection_url = 'oracle+cx_oracle://' + username + ':' + password + '@' + host + '/?service_name=' + service
        engine = create_engine(connection_url)
        conn = engine.connect()
        logger.info('Connection successful !!')
        data = extract_data(conn, query, chunk_size)
        return data
    elif db_type == 'postgres' or db_type == 'aurora-postgres':
        from sqlalchemy import create_engine
        logger.info('Pulling data from PostgreSQL')
        connection_url = 'postgresql://' + username + ':' + password + '@' + host + '/' + database
        engine = create_engine(connection_url)
        conn = engine.connect()
        logger.info('Connection successful !!')
        data = extract_data(conn, query, chunk_size)
        return data


def extract_data(conn=None, query=None, chunk_size=None):
    dfs = []
    record_count = 0

    for chunk in pd.read_sql_query(sql=query, con=conn, chunksize=chunk_size):
        dfs.append(chunk)
        record_count = record_count + chunk.shape[0]
        logger.debug('Records pulled in this batch {}'.format(chunk.shape[0]))
    data = pd.concat(dfs, ignore_index=True)
    logger.info('Total records pulled {}'.format(record_count))

    data.columns = data.columns.str.replace('\W', '_')  # TODO: Revisit this code
    return data



# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: May 29, 2023
# Last modified: June 18, 2023

import numpy as np

default_params = {
    'dtype': [None],
    'output_type': ['array'],
    'random_state': [2],
    'verbose': [4]
}

ElasticNet = {
    'alpha': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0],
    'l1_ratio': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
    'fit_intercept': [True, False],
    'normalize': [True, False],
    'max_iter': np.arange(100, 5000 + 1, 50).tolist(),
    'tol': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'selection': ['cyclic', 'random'],
    'remove_params': ['handle', 'solver']
}

KernelRidge = {
    'kernel': ['additive_chi2', 'cosine', 'laplacian', 'linear', 'poly', 'polynomial', 'rbf', 'sigmoid'],  # 'chi2' is not included as its throwing error
    'gamma': [1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'coef0': [1],
    'degree': [3],
    'remove_params': ['handle', 'alpha', 'kernel_params']
}

KNeighborsRegressor = {
    'algorithm': ['auto'],
    'algo_params': [None],
    'metric': ['euclidean'],
    'metric_params': [None],
    'n_jobs': [None],
    'n_neighbors': [1, 2, 3, 4, 5, 6],
    'p': [1, 2, 3, 4, 5],
    'weights': ['uniform'],
    'remove_params': ['handle']
}

LinearRegression = {
    'algorithm': ['svd', 'eig', 'qr', 'svd-qr', 'svd-jacobi'],
    'fit_intercept': [True, False],
    'normalize': [True, False],
    'remove_params': ['handle']
}

LinearSVR = {
    'penalty': ['l1', 'l2'],
    'loss': ['epsilon_insensitive', 'squared_epsilon_insensitive'],
    'fit_intercept': [True, False],
    'penalized_intercept': [True, False],
    'max_iter': np.arange(100, 5000 + 1, 50).tolist(),
    'linesearch_max_iter': np.arange(100, 1000 + 1, 50).tolist(),
    'lbfgs_memory': [5, 10, 20, 30],
    'C': [1.0],
    'grad_tol': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'change_tol': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'tol': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'epsilon': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1, 0],  # TODO: Revisit this variable in other estimators
    'remove_params': ['handle']
}

Lasso = {
    'alpha': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0],  # TODO: Revisit this variable in other estimators
    'fit_intercept': [True, False],
    'normalize': [True, False],
    'max_iter': np.arange(100, 5000 + 1, 50).tolist(),
    'tol': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'selection': ['cyclic', 'random'],
    'remove_params': ['handle', 'solver']  # solver is throwing error hence removing during hyperparameter tuning
}

MBSGDRegressor = {
    'alpha': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0],
    'l1_ratio': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
    'fit_intercept': [True, False],
    'normalize': [True, False],
    'max_iter': np.arange(100, 5000 + 1, 50).tolist(),
    'tol': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'solver': ['cd', 'qn'],
    'selection': ['cyclic', 'random']
}

RandomForestRegressor = {
    'accuracy_metric': ['r2'],
    'bootstrap': [True],
    'class_weight': [None],
    'criterion': [None],
    'split_criterion': [2], # 0 or 'gini' for gini impurity; 1 or 'entropy' for information gain (entropy); 2 or 'mse' for mean squared error; 4 or 'poisson' for poisson half deviance; 5 or 'gamma' for gamma half deviance; 6 or 'inverse_gaussian' for inverse gaussian deviance
    'max_batch_size': [4096],
    'max_depth': [2, 4, 8, 12, 16],
    'max_features': ['auto', 'sqrt', 'log2'],
    'max_leaf_nodes': [None],
    'max_leaves': [-1],
    'max_samples': [0.7, 0.8, 0.9, 1.0],
    'min_samples_leaf': np.arange(1, 10, 1).tolist(),
    'min_samples_split': np.arange(2, 21, 3).tolist(),
    'min_impurity_decrease': [0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
    'min_impurity_split': [None],
    'min_weight_fraction_leaf': [None],
    'n_bins': [128],
    'n_estimators': [100],
    'n_jobs': [None],
    'n_streams': [4],
    'oob_score': [None],
    'warm_start': [None],
    'remove_params': ['handle']
}

Ridge = {
    'alpha': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0],
    'solver': ['eig', 'svd'],  # 'cd' is not included as its throwing error
    'fit_intercept': [True, False],
    'normalize': [True, False],
    'remove_params': ['handle']
}

SVR = {
    'C': [1.0],
    'kernel': ['linear', 'poly', 'rbf', 'sigmoid'],
    'degree': [1, 2, 3, 4, 5],
    'gamma': ['auto', 'scale'],
    'coef0': [0.0],
    'tol': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1],
    'epsilon': [1e-07, 1e-06, 1e-05, 0.0001, 0.001, 0.01, 0.1, 0],  # TODO: Revisit this variable in other estimators
    'cache_size': [1024.0],
    'class_weight': [None, 'balanced'],
    'max_iter': np.arange(100, 5000 + 1, 50).tolist(),
    'multiclass_strategy': ['ovo', 'ovr'],
    'nochange_steps': [1000],
    'probability': [True, False],
    'remove_params': ['handle']
}




__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"


from bluemist.pipeline.bluemist_pipeline import (get_model_pipeline, clear_all_model_pipelines)




# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: Jun 22, 2022
# Last modified: June 11, 2023

from sklearn.preprocessing import StandardScaler, MaxAbsScaler, MinMaxScaler, RobustScaler

available_scalers = {
    'StandardScaler': StandardScaler,
    'MinMaxScaler': MinMaxScaler,
    'MaxAbsScaler': MaxAbsScaler,
    'RobustScaler': RobustScaler
}


def get_scaler(scaling_strategy):
    return available_scalers.get(scaling_strategy, None)()



import logging
import os
import sys
from logging import config

import sklearn
import cpuinfo
import sklearnex
#import cuml
from sklearnex import patch_sklearn, sklearn_is_patched

#patch_sklearn()

from sklearn import datasets

from bluemist.environment import initialize, check_gpu_brand
# from bluemist.datasource.aws import get_data_from_s3
from bluemist.datasource import get_data_from_filesystem
# from bluemist.pipeline import get_model_pipeline
from bluemist.preprocessing import preprocess_data
from bluemist.regression import train_test_evaluate, deploy_model, get_estimators

# from bluemist.eda import perform_eda
# from bluemist.datasource.database import get_data_from_database

HOME_PATH = os.environ["BLUEMIST_PATH"]
config.fileConfig(HOME_PATH + '/' + 'logging.config')
logger = logging.getLogger("root")

def main():
    # query = 'SELECT * FROM public.auto_mpg'
    # data = get_data_from_database(db_type='postgres', host='postgres.cxh6nuaszc34.us-east-1.rds.amazonaws.com:5432',
    #                        username='postgres', password='adminadmin', database='postgres', query=query, chunk_size=100)
    #
    # query = 'SELECT * FROM public.auto_mpg'
    # data = get_data_from_database(db_type='aurora-postgres', host='aurora-postgres-instance-1.cxh6nuaszc34.us-east-1.rds.amazonaws.com:5432',
    #                        username='postgres', password='adminadmin', database='postgres', query=query, chunk_size=100)
    #
    # query = 'SELECT * FROM auto_mpg'
    # data = get_data_from_database(db_type='mysql', host='mysql.cxh6nuaszc34.us-east-1.rds.amazonaws.com:3306',
    #                        username='admin', password='adminadmin', database='sys', query=query, chunk_size=100)
    #
    # query = 'SELECT * FROM AUTO_MPG'
    # data = get_data_from_database(db_type='aurora-mysql', host='aurora-mysql-instance-1.cxh6nuaszc34.us-east-1.rds.amazonaws.com:3306',
    #                        username='admin', password='adminadmin', database='sys', query=query, chunk_size=100)
    #
    # query = 'SELECT * FROM dbo.auto_mpg'
    # data = get_data_from_database(db_type='mssql', host='mssql.cxh6nuaszc34.us-east-1.rds.amazonaws.com:1433',
    #                        username='admin', password='adminadmin', database='bluemist', query=query, chunk_size=100)
    #
    # query = 'SELECT * FROM auto_mpg'
    # data = get_data_from_database(db_type='mariadb', host='mariadb.cxh6nuaszc34.us-east-1.rds.amazonaws.com:3306',
    #                        username='admin', password='adminadmin', database='innodb', query=query, chunk_size=100)
    #
    #  query = 'SELECT * FROM AUTO_MPG'
    # data = get_data_from_database(db_type='oracle', host='oracle.cxh6nuaszc34.us-east-1.rds.amazonaws.com',
    #                               username='admin', password='adminadmin', service='DATABASE',
    #                               oracle_instant_client_path='/home/shashank-agrawal/Desktop/instantclient_21_6',
    #                               query=query, chunk_size=100)

    initialize(enable_acceleration_extensions=True)
    # print('get_estimators', get_estimators())
    info = cpuinfo.get_cpu_info()
    print(info)
    print(sklearn_is_patched())


    # data = datasets.load_diabetes(as_frame=True)
    # data = get_data_from_filesystem('datasets/auto-mpg/auto-mpg.csv')

    data = get_data_from_filesystem(
        'https://raw.githubusercontent.com/plotly/datasets/3aa08e58607d1f36159efc4cca9d0d073bbf57bb/auto-mpg.csv')

    # perform_eda(data.frame, target_variable='target', provider='autoviz')

    X_train, X_test, y_train, y_test = preprocess_data(data,
                                                       target_variable='mpg',
                                                       test_size=0.25,
                                                       data_scaling_strategy='StandardScaler',
                                                       categorical_features=['model_year'],
                                                       categorical_encoder='OneHotEncoder',
                                                       drop_categories_one_hot_encoder='first')
    train_test_evaluate(X_train, X_test, y_train, y_test)
    deploy_model(estimator_name='LinearSVR')
    # # # pipeline = get_model_pipeline('LarsCV')
    # # print(pipeline.get_params)
    # deploy_model(estimator_name='LarsCV', host='localhost', port=8000)


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    main()




__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"


__all__ = [
    'datasource',
    'eda',
    'preprocessing',
    'regression',
    'classification'
]




__author__ = "Shashank Agrawal"
__license__ = "MIT"
__version__ = "0.1.1"
__email__ = "dew@bluemist-ai.one"

import logging
import os
from logging import config

from pandas_profiling import ProfileReport

BLUEMIST_PATH = os.getenv("BLUEMIST_PATH")
EDA_ARTIFACTS_PATH = BLUEMIST_PATH + '/' + 'artifacts/eda'

config.fileConfig(BLUEMIST_PATH + '/' + 'logging.config')
logger = logging.getLogger("bluemist")


def perform_eda(data,
                provider='pandas-profiling',
                sample_size=10000,
                data_randomizer=2):
    """
        Performs Exploratory Data Analysis (EDA)

        data: pandas dataframe
            Dataframe for exploratory data analysis
        provider : {'pandas-profiling', 'sweetviz', 'dtale'}, default='pandas-profiling'
            Library provider for exploratory data analysis
        sample_size: str, default=10000
            Number of rows to return from dataframe. ``None`` to perform eda on the complete dataset which can be slower if dataset has large number of rows and columns
        data_randomizer: int, default=None
            Controls the data split. Provide a value to reproduce the same split.

        Examples
        ---------
        *EDA using Pandas Profiling*

        .. raw:: html
           :file: ../../code_samples/quickstarts/eda/eda_pandas-profiling.html

        *EDA using SweetVIZ*

        .. raw:: html
           :file: ../../code_samples/quickstarts/eda/eda_sweetviz.html

        *EDA using D-TALE*

        .. raw:: html
           :file: ../../code_samples/quickstarts/eda/eda_dtale.html

    """

    if sample_size is not None and data.shape[0] >= sample_size:
        data = data.sample(n=sample_size, random_state=data_randomizer)

    output_provider = EDA_ARTIFACTS_PATH + '/' + provider
    output_file = output_provider + '.html'

    valid_providers = ['pandas-profiling', 'sweetviz', 'dtale', 'autoviz']
    if provider in valid_providers:
        logger.info('Peforming EDA using :: {}'.format(provider))
        if provider == 'pandas-profiling':
            logger.info('Output file :: {}'.format(output_file))
            print('Output file :: {}'.format(output_file))
            print('Output file will be opened in the browser after analysis is completed !!')
            profile = ProfileReport(data, explorative=True)
            profile.to_file(output_file=output_file, silent=False)
        elif provider == 'sweetviz':
            import sweetviz as sv
            logger.info('Output file :: {}'.format(output_file))
            print('Output file :: {}'.format(output_file))
            print('Output file will be opened in the browser after analysis is completed !!')
            sweetviz_report = sv.analyze(data)
            sweetviz_report.show_html(output_file)
        elif provider == 'dtale':
            import dtale
            print('Opening dtale UI on the browser...')
            d = dtale.show(data, subprocess=False, reaper_on=True)
            d.open_browser()
    else:
        print('Invalid provider, valid providers are :: {}'.format(valid_providers))
        logger.info('Invalid provider, valid providers are :: {}'.format(valid_providers))



# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: Sep 6, 2022
# Last modified: June 19, 2023


from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import PowerTransformer

from bluemist.utils.scaler import get_scaler, available_scalers


def build_numeric_transformer_pipeline(**kwargs):
    preprocessing_steps = []

    imputer_strategy = kwargs.get('numeric_imputer_strategy')
    scaler = kwargs.get('data_scaling_strategy')
    numeric_constant_value = kwargs.get('numeric_constant_value')
    data_tranformation_strategy = kwargs.get('data_tranformation_strategy')

    # handling missing values
    if imputer_strategy is not None and imputer_strategy in ['mean', 'median', 'most_frequent', 'constant']:
        if imputer_strategy == 'constant':
            imputer_step = ('imputer', SimpleImputer(strategy=imputer_strategy, fill_value=numeric_constant_value))
        else:
            imputer_step = ('imputer', SimpleImputer(strategy=imputer_strategy))
        preprocessing_steps.append(imputer_step)
    elif imputer_strategy is not None:
        raise ValueError('Invalid imputer_strategy value passed : ', imputer_strategy)

    # data scaling
    if scaler is not None and scaler in available_scalers:
        scaler_step = ('scaler', get_scaler(scaler))
        preprocessing_steps.append(scaler_step)
    elif scaler is not None:
        raise ValueError('Invalid scaler value passed : ', scaler)

    # data transformation
    if data_tranformation_strategy is not None and data_tranformation_strategy in ['auto', 'yeo-johnson', 'box-cox']:
        transformer_step = tuple()
        if data_tranformation_strategy == 'yeo-johnson':
            transformer_step = ('transformer', PowerTransformer(method='yeo-johnson'))
        elif data_tranformation_strategy == 'box-cox':
            transformer_step = ('transformer', PowerTransformer(method='box-cox'))
        preprocessing_steps.append(transformer_step)

    numeric_transformer = Pipeline(steps=preprocessing_steps)
    return numeric_transformer




# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: Jun 11, 2023
# Last modified: Jun 19, 2023

GPU_BRAND_NVIDIA = 'NVIDIA'
GPU_BRAND_INTEL = 'Intel'
CPU_BRAND_INTEL = 'GenuineIntel'

GPU_ACCELERATION_NVIDIA = 'NVIDIA GPU Acceleration'
CPU_ACCELERATION_INTEL = 'Intel CPU Acceleration'



# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created:  Feb 19, 2023
# Last modified: June 19, 2023

import os
from jinja2 import Template

BLUEMIST_PATH = os.environ["BLUEMIST_PATH"]

class_template = """import nest_asyncio
import pandas as pd
import joblib
import uvicorn
from pydantic import BaseModel
from fastapi import FastAPI
from pyngrok import ngrok
import os


class request_body(BaseModel):
    {%+ for column, data_type in initial_column_metadata -%}
        {{ column }}: np.{{ data_type.name }}
    {%+ endfor -%}

"""

func_template = """

app = FastAPI(debug=True)

BLUEMIST_PATH = os.environ["BLUEMIST_PATH"]
preprocessor = joblib.load(BLUEMIST_PATH + '/' + 'artifacts/preprocessor/preprocessor.joblib')
pipeline = joblib.load(BLUEMIST_PATH + '/' + 'artifacts/models/{{ estimator_name }}.joblib')


@app.post('/predict')
def predict(data: request_body):
    # Making the data in a form suitable for prediction
    input_data = [[
        {%+ for column, _ in initial_column_metadata -%}
            data.{{ column }},
        {%+ endfor -%}
        ]]

    input_df = pd.DataFrame(input_data, columns=[
        {%+ for column, _ in initial_column_metadata -%}
            '{{ column }}',
        {%+ endfor -%}
        ])

    df_to_predict = pd.DataFrame(preprocessor.transform(input_df), columns=[
        {%+ for column in encoded_column_metadata -%}
            '{{ column }}',
        {%+ endfor -%}
        ])

    # Predicting the Class
    prediction = pipeline.predict(df_to_predict)

    # Return the Result
    return {'predicted_{{ target_variable }}': prediction[0]}


def start_api_server(host='localhost', port=8000):
    ngrok_tunnel = ngrok.connect(port)
    ngrok_tunnel
    nest_asyncio.apply()
    uvicorn.run(app, host=host, port=port)

"""


def generate_api_code(estimator_name, initial_column_metadata, encoded_column_metadata, target_variable):

    template = Template(class_template)
    class_code = template.render(initial_column_metadata=initial_column_metadata)

    class_code = class_code.replace('np.int64', 'int')\
        .replace('np.float64', 'float')\
        .replace('np.object', 'str')

    template = Template(func_template)
    func_code = template.render(initial_column_metadata=initial_column_metadata,
                                encoded_column_metadata=encoded_column_metadata, estimator_name=estimator_name,
                                target_variable=target_variable)

    with open(BLUEMIST_PATH + '/' + 'artifacts/api/predict.py', 'w') as f:
        f.truncate()
        f.write(class_code)
        f.write(func_code)






from setuptools import setup, find_packages

with open("requirements.txt") as file:
    required_packages = file.read().splitlines()

with open("requirements-optional.txt") as file:
    optional_packages = file.read().splitlines()

setup(
    packages=find_packages(include=["bluemist*"]),
    install_requires=required_packages,
    extras_require={
        "complete": [optional_packages]
    }
)



# Author: Shashank Agrawal
# License: MIT
# Version: 0.1.2
# Email: dew@bluemist-ai.one
# Created: Sep 6, 2022
# Last modified: May 29, 2023

import sklearn
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder, OneHotEncoder

from bluemist import environment


def build_categorical_transformer_pipeline(**kwargs):
    transformer_steps = []

    imputer_strategy = kwargs.get('categorical_imputer_strategy')
    categorical_constant_value = kwargs.get('categorical_constant_value')
    categorical_encoder = kwargs.get('categorical_encoder')
    drop_categories = kwargs.get('drop_categories_one_hot_encoder')
    handle_unknown = kwargs.get('handle_unknown_one_hot_encoder')

    if imputer_strategy is not None and imputer_strategy in ['most_frequent', 'constant']:
        if imputer_strategy == 'constant':
            imputer_step = ('imputer', SimpleImputer(strategy=imputer_strategy, fill_value=categorical_constant_value))
        else:
            imputer_step = ('imputer', SimpleImputer(strategy=imputer_strategy))
        transformer_steps.append(imputer_step)
    else:
        raise ValueError('Invalid imputer_strategy value passed: {}'.format(imputer_strategy))

    if categorical_encoder is not None and categorical_encoder in ['OneHotEncoder', 'OrdinalEncoder']:
        encoder_step = tuple()
        if categorical_encoder == 'LabelEncoder':
            encoder_step = ('label_encoder', LabelEncoder())
        elif categorical_encoder == 'OrdinalEncoder':
            encoder_step = ('ordinal_encoder', OrdinalEncoder())
        elif categorical_encoder == 'OneHotEncoder':
            encoder_class = OneHotEncoder
            kwargs = {}
            if drop_categories is not None:
                kwargs['drop'] = drop_categories
            if handle_unknown is not None:
                kwargs['handle_unknown'] = handle_unknown

            encoder_step = ('one_hot_encoder', encoder_class(**kwargs))

        transformer_steps.append(encoder_step)
    else:
        raise ValueError('Invalid categorical_encoder value passed : ', categorical_encoder)

    categorical_transformer = Pipeline(steps=transformer_steps)
    return categorical_transformer
